import typing as t

from geodefi.globals import ID_TYPE, DEPOSIT_SIZE
from geodefi.utils import get_key, validate_deposit_data_file


from .id import Id
from .validator import Validator
from .beacon import Beacon


class Pool(Id):
    def __init__(self, beacon: Beacon, *args, **kwargs):
        """
        Initializes a Pool object.

        Parameters:
            beacon: An initialized Beacon object.
            *args: variable length argument list.
            **kwargs: arbitrary keyword arguments.

        Returns:
        None
        """
        # The type of the Pool is 5.
        super().__init__(*args, **kwargs, type_=ID_TYPE(5))
        # The Pool object is associated with a Beacon object.
        self.beacon: Beacon = beacon

    @property
    def initiated(self):
        return self._read_uint("initiated")

    @property
    def maintainer(self):
        return self._read_address("maintainer")

    @property
    def yieldReceiver(self):
        return self._read_address("yieldReceiver")

    @property
    def surplus(self):
        return self._read_uint("surplus")

    @property
    def secured(self):
        return self._read_uint("secured")

    @property
    def middlewares_list(self):
        lenmiddlewares = self.middlewares_len
        ints: t.List = []
        for i in range(lenmiddlewares):
            ints.append(self._read_address_array("middlewares", i))
        return ints

    @property
    def middlewares_len(self):
        return self._read_uint("middlewares")

    def middlewares(self, index):
        return self._read_address_array("middlewares", index)

    @property
    def private(self):
        return self._read_uint("private") == 1

    @property
    def whitelist(self):
        return self._read_address("whitelist")

    @property
    def withdrawalCredential(self):
        return self._read_bytes("withdrawalCredential", is_hex=True)

    @property
    def withdrawalContract(self):
        return self._read_address("withdrawalContract")

    @property
    def liquidityPool(self):
        return self._read_address("liquidityPool")

    @property
    def feeSwitch(self):
        return self._read_uint("feeSwitch")

    @property
    def priorFee(self):
        return self._read_uint("priorFee")

    @property
    def fee(self):
        return self._read_uint("fee")

    @property
    def wallet(self):
        return self._read_uint("wallet")

    @property
    def validatorsList(self):
        lenvalidators = self.validatorsLen
        vals: t.List = []
        for i in range(lenvalidators):
            vals.append(self._read_bytes_array("validators", index=i))
        return vals

    @property
    def validatorsLen(self):
        return self._read_uint("validators")

    @property
    def fallbackOperator(self):
        return self._read_uint("fallbackOperator")

    @property
    def fallbackThreshold(self):
        return self._read_uint("fallbackThreshold")

    def validators(self, index):
        return Validator(
            network=self.network,
            beacon=self.beacon,
            w3=self.w3,
            portal=self.portal,
            pk=self._read_bytes_array("validators", index=index),
        )

    def allowance(self, operator: int):
        return self._read_uint(get_key(id_=operator, key="allowance"))

    def proposedValidators(self, operator: int):
        return self._read_uint(get_key(id_=operator, key="proposedValidators"))

    def activeValidators(self, operator: int):
        return self._read_uint(get_key(id_=operator, key="activeValidators"))

    def alienValidators(self, operator: int):
        return self._read_uint(get_key(id_=operator, key="alienValidators"))

    def prepareProposeStake(self, deposit_data_path: str):
        """
        This method prepares for a validator proposal. It reads deposit data from a file path,
        validates it, and returns the public keys and signature 1s for the proposal stake.

        Args:
        - deposit_data_path (str): The file path of the deposit data.

        Returns:
        - pubkeys (List[bytes]): A list of public keys as bytes objects.
        - sig1s (List[bytes]): A list of signature 1s as bytes objects.
        """
        deposit_data = validate_deposit_data_file(
            deposit_data_path=deposit_data_path,
            amount=DEPOSIT_SIZE.PROPOSAL,
            network=self.network,
            credential=self.withdrawalCredential[2:],
        )
        pubkeys = [bytes.fromhex(deposit["pubkey"]) for deposit in deposit_data]
        sig1s = [
            bytes.fromhex(deposit["signature"]) for deposit in deposit_data
        ]
        return pubkeys, sig1s

    def prepareStake(self, deposit_data_path: str):
        """
        This function prepares a beacon stake by taking the path to a deposit data file as input.
        The deposit data file is validated by checking that it contains the required amount of Ether
        for the Beacon chain, that it is meant for the specified network, and that it is associated
        with the withdrawal credentials of the current operator.

        Once the deposit data has been validated, the function extracts the public keys and
        signatures from the deposit data and returns them as a tuple.

        Args:
            deposit_data_path (str): The path to the deposit data file.

        Returns:
            Tuple of lists: A tuple containing two lists, the public keys and signatures extracted
            from the deposit data file.
        """

        deposit_data = validate_deposit_data_file(
            deposit_data_path=deposit_data_path,
            amount=DEPOSIT_SIZE.STAKE,
            network=self.network,
            credential=self.withdrawalCredential[2:],
        )
        pubkeys = [bytes.fromhex(deposit["pubkey"]) for deposit in deposit_data]
        sig31s = [
            bytes.fromhex(deposit["signature"]) for deposit in deposit_data
        ]
        return pubkeys, sig31s



class PythonVersionError(Exception):
    "Current version is < 3.7 or > 3.10"


class UnknownChainError(Exception):
    "Provided RPC points to an unknown chainId"


class UnexpectedResponseError(Exception):
    "Https call resulted with an error code (non 2xx)"


class HTTPRequestError(Exception):
    "RESPONSE IS NOT CODE 200"


class MaxAttemptError(Exception):
    "MULTIPLE ATTEMPT MADE BUT CALL FAILED"


class DepositSizeError(Exception):
    "DEPOSIT SIZE IS NOT CORRECT !(1/31)"


class WithdrawalCredentialError(Exception):
    "INCORRECT withdrawal_credentials"


class GenesisForkError(Exception):
    "INCORRECT fork_version"


class NetworkNameError(Exception):
    "INCORRECT network_name"


class DepositDataError(Exception):
    "CAN NOT VERIFY DEPOSIT DATA"



from os import path
import json
import typing as t
from geodefi.globals import (
    Network,
    DEPOSIT_SIZE,
    DEPOSIT_NETWORK_NAME,
    GENESIS_FORK_VERSION,
)
from geodefi.exceptions import (
    DepositSizeError,
    WithdrawalCredentialError,
    GenesisForkError,
    NetworkNameError,
    DepositDataError,
)

from .validate import validate_deposit


def resource_path(relative_path: str) -> str:
    """
    Get the absolute path to a resource in a manner friendly to PyInstaller.
    PyInstaller creates a temp folder and stores path in _MEIPASS which this function swaps
    into a resource path so it is available both when building binaries and running natively.
    """
    try:
        base_path = sys._MEIPASS  # type: ignore
    except Exception:
        base_path = path.abspath(".")
    return path.join(base_path, relative_path)


def _get_deposit_data(deposit_data_path: str):
    file_path = resource_path(deposit_data_path)
    with open(file_path, "r") as file:
        a = file.read()
    return json.loads(a)


def validate_deposit_data_file(
    deposit_data_path: str,
    amount: DEPOSIT_SIZE,
    credential: str,
    network: Network,
):
    deposit_data: t.List[t.Dict] = _get_deposit_data(deposit_data_path)
    for deposit in deposit_data:
        if deposit["amount"] != amount.value:
            raise DepositSizeError
        if deposit["withdrawal_credentials"] != credential:
            raise WithdrawalCredentialError
        if (
            bytes.fromhex(deposit["fork_version"])
            != GENESIS_FORK_VERSION[network.value]
        ):
            raise GenesisForkError
        if deposit["network_name"] != DEPOSIT_NETWORK_NAME[network.value]:
            raise NetworkNameError
        if not validate_deposit(deposit):
            raise DepositDataError
    return deposit_data



import sys
import os
import pytest

sys.path.append(os.getcwd())

try:
    import geodefi
    from geodefi.utils.crypto import keccak256, solidity_keccak256
except ImportError:
    raise


@pytest.mark.parametrize(
    "input_data, expected_hash",
    [
        (
            b"Hello, World!",
            b"\xac\xaf2\x89\xd7\xb6\x01\xcb\xd1\x14\xfb6\xc4\xd2\x9c\x85\xbb\xfd^\x13?\x14\xcb5\\?\xd8\xd9\x93g\x96O",
        ),
        (
            b"12345",
            b"\x18A\xd6S\xf9\xc4\xed\xda\x9df\xa7\xe7s{9v=k\xd4\x0fV\x9a>\xc6\x85\x9d3\x05\xb7#\x10\xe6",
        ),
        (
            b"",
            b"\xc5\xd2F\x01\x86\xf7#<\x92~}\xb2\xdc\xc7\x03\xc0\xe5\x00\xb6S\xca\x82';{\xfa\xd8\x04]\x85\xa4p",
        ),
    ],
)
def test_keccak256(input_data, expected_hash):

    result = keccak256(input_data)
    assert result == expected_hash


@pytest.mark.parametrize(
    "types, input_data, expected_hash",
    [
        (
            ["uint256"],
            [123],
            b"Ui\x04G\x19\xa1\xec;\x04\xd0\xaf\xa9\xe7\xa51\x0c|\x04s3\x1d\x13\xdc\x9f\xaf\xe1C\xb2\xc4\xe8\x14\x8a",
        ),
        (
            ["bool", "address", "uint256"],
            [True, "0xAb8483F64d9C6d1EcF9b849Ae677dD3315835cb2", 98765],
            b"2_\xd8\xfc\xb4v\x84\x18,e\xe7\xb6\xce\x9e\xa5\x8e2\xaf1d\xb1\x855\n\x08\xea\x1bX\x0b\xb3\xc0\x00",
        ),
        (
            ["bool", "bool"],
            [True, False],
            b"b\x8b\xf3YgG\xd23\xf1\xe6S3Ep\x00f\xbfE\x8f\xa4\x8d\xae\xda\xf0J{\xe6\xc3\x92\x90$v",
        ),
    ],
)
def test_solidity_keccak256(types, input_data, expected_hash):

    result = solidity_keccak256(types, input_data)
    assert result == expected_hash



from dataclasses import asdict, dataclass, fields, field as dataclass_field
import json
import os
from py_ecc.bls import G2ProofOfPossession as bls
from secrets import randbits
from typing import Any, Dict, Union
from unicodedata import normalize
from uuid import uuid4

from geodefi.utils.bls.crypto import (
    AES_128_CTR,
    PBKDF2,
    scrypt,
    SHA256,
)

UNICODE_CONTROL_CHARS = list(range(0x00, 0x20)) + list(range(0x7F, 0xA0))


hexdigits = set("0123456789abcdef")


def encode_bytes(
    obj: Union[str, Dict[str, Any]]
) -> Union[bytes, str, Dict[str, Any]]:
    """
    Recursively encodes objects that contain hexstrings into objects that contain bytes.
    """
    if isinstance(obj, str) and all(c in hexdigits for c in obj):
        return bytes.fromhex(obj)
    elif isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = encode_bytes(value)
    return obj


class BytesDataclass:
    """
    BytesDataClasses are DataClass objects that automatically encode hexstrings into bytes,
    and have an `as_json` function that encodes bytes back into hexstrings.
    """

    def __post_init__(self) -> None:
        for field in fields(self):
            if field.type in (bytes, Dict[str, Any]):
                # Convert hexstring to bytes
                self.__setattr__(
                    field.name, encode_bytes(self.__getattribute__(field.name))
                )

    def as_json(self) -> str:
        return json.dumps(asdict(self), default=lambda x: x.hex())


@dataclass
class KeystoreModule(BytesDataclass):
    function: str = ""
    params: Dict[str, Any] = dataclass_field(default_factory=dict)
    message: bytes = bytes()


@dataclass
class KeystoreCrypto(BytesDataclass):
    kdf: KeystoreModule = KeystoreModule()
    checksum: KeystoreModule = KeystoreModule()
    cipher: KeystoreModule = KeystoreModule()

    @classmethod
    def from_json(cls, json_dict: Dict[Any, Any]) -> "KeystoreCrypto":
        kdf = KeystoreModule(**json_dict["kdf"])
        checksum = KeystoreModule(**json_dict["checksum"])
        cipher = KeystoreModule(**json_dict["cipher"])
        return cls(kdf=kdf, checksum=checksum, cipher=cipher)


@dataclass
class Keystore(BytesDataclass):
    """
    Implement an EIP 2335-compliant keystore. A keystore is a JSON file that
    stores an encrypted version of a private key under a user-supplied password.

    Ref: https://github.com/ethereum/EIPs/blob/master/EIPS/eip-2335.md
    """

    crypto: KeystoreCrypto = KeystoreCrypto()
    description: str = ""
    pubkey: str = ""
    path: str = ""
    uuid: str = ""
    version: int = 4

    def kdf(self, **kwargs: Any) -> bytes:
        return (
            scrypt(**kwargs)
            if "scrypt" in self.crypto.kdf.function
            else PBKDF2(**kwargs)
        )

    def save(self, filefolder: str) -> None:
        """
        Save self as a JSON keystore.
        """
        with open(filefolder, "w") as f:
            f.write(self.as_json())
        if os.name == "posix":
            os.chmod(filefolder, int("440", 8))  # Read for owner & group

    @classmethod
    def from_json(cls, json_dict: Dict[Any, Any]) -> "Keystore":
        crypto = KeystoreCrypto.from_json(json_dict["crypto"])
        path = json_dict["path"]
        uuid = json_dict["uuid"]
        version = json_dict["version"]
        description = json_dict.get("description", "")
        pubkey = json_dict.get("pubkey", "")
        return cls(
            crypto=crypto,
            description=description,
            pubkey=pubkey,
            path=path,
            uuid=uuid,
            version=version,
        )

    @classmethod
    def from_file(cls, path: str) -> "Keystore":
        with open(path, "r") as f:
            return cls.from_json(json.load(f))

    @staticmethod
    def _process_password(password: str) -> bytes:
        """
        Encode password as NFKD UTF-8 as per:
        https://github.com/ethereum/EIPs/blob/master/EIPS/eip-2335.md#password-requirements
        """
        password = normalize("NFKD", password)
        password = "".join(
            c for c in password if ord(c) not in UNICODE_CONTROL_CHARS
        )
        return password.encode("UTF-8")

    @classmethod
    def encrypt(
        cls,
        *,
        secret: bytes,
        password: str,
        path: str = "",
        kdf_salt: bytes = randbits(256).to_bytes(32, "big"),
        aes_iv: bytes = randbits(128).to_bytes(16, "big")
    ) -> "Keystore":
        """
        Encrypt a secret (BLS SK) as an EIP 2335 Keystore.
        """
        keystore = cls()
        keystore.uuid = str(uuid4())
        keystore.crypto.kdf.params["salt"] = kdf_salt
        decryption_key = keystore.kdf(
            password=cls._process_password(password),
            **keystore.crypto.kdf.params
        )
        keystore.crypto.cipher.params["iv"] = aes_iv
        cipher = AES_128_CTR(
            key=decryption_key[:16], **keystore.crypto.cipher.params
        )
        keystore.crypto.cipher.message = cipher.encrypt(secret)
        keystore.crypto.checksum.message = SHA256(
            decryption_key[16:32] + keystore.crypto.cipher.message
        )
        keystore.pubkey = bls.SkToPk(int.from_bytes(secret, "big")).hex()
        keystore.path = path
        return keystore

    def decrypt(self, password: str) -> bytes:
        """
        Retrieve the secret (BLS SK) from the self keystore by decrypting it with `password`
        """
        decryption_key = self.kdf(
            password=self._process_password(password), **self.crypto.kdf.params
        )
        if (
            SHA256(decryption_key[16:32] + self.crypto.cipher.message)
            != self.crypto.checksum.message
        ):
            raise ValueError("Checksum message error")

        cipher = AES_128_CTR(
            key=decryption_key[:16], **self.crypto.cipher.params
        )
        return cipher.decrypt(self.crypto.cipher.message)


@dataclass
class Pbkdf2Keystore(Keystore):
    crypto: KeystoreCrypto = KeystoreCrypto(
        kdf=KeystoreModule(
            function="pbkdf2",
            params={"c": 2**18, "dklen": 32, "prf": "hmac-sha256"},
        ),
        checksum=KeystoreModule(
            function="sha256",
        ),
        cipher=KeystoreModule(
            function="aes-128-ctr",
        ),
    )


@dataclass
class ScryptKeystore(Keystore):
    crypto: KeystoreCrypto = KeystoreCrypto(
        kdf=KeystoreModule(
            function="scrypt",
            params={
                "dklen": 32,
                "n": 2**18,
                "r": 8,
                "p": 1,
            },
        ),
        checksum=KeystoreModule(
            function="sha256",
        ),
        cipher=KeystoreModule(
            function="aes-128-ctr",
        ),
    )



from eth_typing import (
    BLSPubkey,
    BLSSignature,
)
from typing import Any, Dict, List, Optional
from py_ecc.bls import G2ProofOfPossession as bls
from Crypto.Hash import SHA256 as _sha256

from geodefi.globals import MIN_DEPOSIT_AMOUNT, MAX_DEPOSIT_AMOUNT

from .serialize import (
    DepositMessage,
    compute_deposit_domain,
    compute_signing_root,
    DepositData,
)

from .staking_deposit import from_mnemonic, get_mnemonic


def SHA256(x: bytes) -> bytes:
    return _sha256.new(x).digest()


def validate_deposit(deposit_data_dict: Dict[str, Any]) -> bool:
    """
    Checks whether a deposit is valid based on the staking deposit rules.
    https://github.com/ethereum/consensus-specs/blob/dev/specs/phase0/beacon-chain.md#deposits
    """
    pubkey = BLSPubkey(bytes.fromhex(deposit_data_dict["pubkey"]))
    withdrawal_credentials = bytes.fromhex(
        deposit_data_dict["withdrawal_credentials"]
    )
    amount = deposit_data_dict["amount"]
    signature = BLSSignature(bytes.fromhex(deposit_data_dict["signature"]))
    deposit_message_root = bytes.fromhex(deposit_data_dict["deposit_data_root"])
    fork_version = bytes.fromhex(deposit_data_dict["fork_version"])

    # Verify pubkey
    if len(pubkey) != 48:
        return False

    # Verify deposit signature && pubkey
    deposit_message = DepositMessage(
        pubkey=pubkey,
        withdrawal_credentials=withdrawal_credentials,
        amount=amount,
    )
    domain = compute_deposit_domain(fork_version)
    signing_root = compute_signing_root(deposit_message, domain)
    if not bls.Verify(pubkey, signing_root, signature):
        return False

    # Verify Deposit Root
    signed_deposit = DepositData(
        pubkey=pubkey,
        withdrawal_credentials=withdrawal_credentials,
        amount=amount,
        signature=signature,
    )
    return signed_deposit.hash_tree_root == deposit_message_root


def validate_parameters(
    pubkey, withdrawal_credentials, amount, signature, fork_version
) -> bool:
    """
    :param: pubkey (str) ==> len-96 "No 0x-prefix" public key.
    :param: withdrawal_credentials (str) ==>  len-96 "No 0x-prefix"
    :param: amount (int) ==> 31000000000 for 31 and 1000000000 for 1.
    :param: signature (str) ==>  len-192 "No 0x-prefix"
    :param: fork_version (bytes) ==>  fork_version

    return True if the parameters can be validated, else otherwise.
    """

    # pubkey
    pubkey = BLSPubkey(bytes.fromhex(pubkey))

    # withdrawal_credentials
    withdrawal_credentials = bytes.fromhex(withdrawal_credentials)

    # signature
    signature = bytes.fromhex(signature)

    # Form deposit masage
    deposit_message = DepositMessage(
        pubkey=pubkey,
        withdrawal_credentials=withdrawal_credentials,
        amount=amount,
    )

    domain = compute_deposit_domain(fork_version)
    signing_root = compute_signing_root(deposit_message, domain)

    return bls.Verify(pubkey, signing_root, signature)


def createDepositData(
    mnemonic: Optional[str],
    mnemonic_password: str,
    start_index: int,
    num_keys: int,
    withdrawal_address: str,
) -> List[Dict]:
    """
    Create deposit data for a blockchain application.

    Args:
        start_index (int): The starting index for key generation.
        mnemonic (Optional[str]): A mnemonic (seed phrase) for key generation.
        mnemonic_password (str): Password for unlocking the mnemonic.
        num_keys (int): The number of keys to generate.

    Returns:
        List[Dict]: A list of dictionaries representing deposit data.

    This function creates deposit data for a blockchain application. It can either use an existing mnemonic or generate a new one if none is provided. It then generates a specified number of keys from the mnemonic and password and writes deposit data to a file.

    If `mnemonic` is not provided, a new mnemonic is generated, displayed to the user, and the user is prompted to record it securely.

    The `from_mnemonic` function is called to generate keys from the mnemonic and password.

    The `write_deposit_data` function is called to save the generated deposit data to a file.
    """

    # TODO chain stable
    chain_id = 5

    bytes_address = bytes.fromhex(withdrawal_address[2:])

    if mnemonic is None:
        # Generate a new mnemonic if not provided
        mnemonic = get_mnemonic()

    # Generate keys from the mnemonic and password
    data_dict, keystore_dict = from_mnemonic(
        mnemonic,
        mnemonic_password,
        chain_id,
        start_index,
        num_keys,
        bytes_address,
    )

    return {
        "mnemonic": mnemonic,
        "deposit_data": data_dict,
        "keystore_data": keystore_dict,
    }



from geodefi.utils.merkle import StandartMerkleTree
from geodefi import Geode
from dotenv import dotenv_values
import sys
import os
import pytest
import inspect
from eth_abi import encode

sys.path.append(os.getcwd())


env = dotenv_values(".env")

#################################################
####### StandartMerkleTree ######################
#################################################


def test_tree_initilize():

    tree = StandartMerkleTree()

    assert tree._tree == []
    assert tree._generated == False
    assert tree._depth == 0
    assert tree._numberOfNodes == 0
    assert tree._sizeOfTree == 0
    assert tree._sort == True


@pytest.mark.parametrize(
    "leaves, depht, numberOfNodes, sizeOfTree",
    [
        ([[0]], 1, 1, 2),
        ([[i] for i in range(16)], 5, 16, 32),
        ([[i] for i in range(31)], 6, 31, 64),
        ([[i] for i in range(157)], 9, 157, 512),
    ],
)
def test_tree_generate(leaves, depht, numberOfNodes, sizeOfTree):

    standartMerkleTree = StandartMerkleTree()

    types = ["int"]

    standartMerkleTree.of(leaves, types)

    # standart merkle tree

    assert standartMerkleTree._generated == True
    assert standartMerkleTree._sort == True

    # tests
    assert standartMerkleTree._depth == depht
    assert standartMerkleTree._numberOfNodes == numberOfNodes
    assert standartMerkleTree._sizeOfTree == sizeOfTree


@pytest.mark.parametrize(
    "leaves, types, root",
    [
        # Test case 1: Integer leaves and root
        (
            [[0]],
            ["int"],
            "510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9",
        ),
        (
            [[i] for i in range(16)],
            ["int"],
            "c6b70ea22c9cb271ba84ea37f14787964259e6960b35ae8ddf51f9bd0e3176a3",
        ),
        (
            [[i] for i in range(31)],
            ["int"],
            "aef0f48d8105f948f9c4c44648fb729659fab531b4cb4a3fe3945a22b5f2e2ee",
        ),
        (
            [[i] for i in range(157)],
            ["int"],
            "a0ec4dd2ba6eb5e1578efd8516284a99d17aa87e70d9e36b8afe4a70d2046b4d",
        ),
        # Test case 2: Bytes leaves and root
        (
            [[b"hello"]],
            ["bytes"],
            "1df03627ad4486d6e8e7be19998ae6e769a905b90c5cfc2519dfe035ab543584",
        ),
        (
            [[b"hello"], [b"world"]],
            ["bytes"],
            "7c8b7867dc71b310336f859cc3d8aa46fc1bb5cfec6b95d0097efa0eb6d32bfc",
        ),
        (
            [[b"test", b"merkle", b"tree"]],
            ["bytes", "bytes", "bytes"],
            "f68251b4a7d6d99ef3b8b9f16f1ade23dc084ad7b8f69e9c491cd018e978c169",
        ),
        # Test case 3: Mix of different types and root
        (
            [
                [
                    1,
                    "0xd3cda913deb6f67967b99d67acdfa1712c293601",
                    b"bytesdata",
                    True,
                ]
            ],
            ["int", "address", "bytes", "bool"],
            "97c424a58930c023d603a0912b6a5a385c3203600e6416a7fcfc5c1ce63b1bf0",
        ),
        (
            [
                [
                    i,
                    f"0xd3cda913deb6f67967b99d67acdfa1712c29360{i}",
                    b"bytesdata" + bytes([i]),
                    bool(i % 2 == 0),
                ]
                for i in range(5)
            ],
            ["int", "address", "bytes", "bool"],
            "328e990b1d6f83dfdb9bbced33fdd8d9e02a2e492d9e04989a0a9a6e758ef593",
        ),
        # Test case 4: Only boolean leaves and root
        (
            [[True]],
            ["bool"],
            "b5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22",
        ),
        (
            [[False, True]],
            ["bool", "bool"],
            "0eb5be412f275a18f6e4d622aee4ff40b21467c926224771b782d4c095d1444b",
        ),
        # Test case 5: Only address leaves and root
        (
            [["0xd3cda913deb6f67967b99d67acdfa1712c293601"]],
            ["address"],
            "0c4f91a58661ab3c69dbe0fe58d1ecea2598699df92381a8681760559123de27",
        ),
        # Test case 6: Mix of integers and bytes and root
        (
            [[1, "0xd3cda913deb6f67967b99d67acdfa1712c293601", b"helloworld"]],
            ["int", "address", "bytes"],
            "5afcd76f065f5e4a55528ac105a6c99bcc81467f99d896185e6cafecac5c2dd1",
        ),
        # Test case 7: Only bytes leaves and root
        (
            [[b"bytesdata"]],
            ["bytes"],
            "818c0af0faad90afaffdbc728e419c819e7b9b3f2ce508a92d778739a82712c6",
        ),
        # Test case 8: Mix of integers and booleans and root
        (
            [[1, False], [42, True]],
            ["int", "bool"],
            "29121ab9e52f223f2f465405ba61deb61bf14d96a96c048dda8d1b117ea89c04",
        ),
        # Test case 9: Mix of bytes and booleans and root
        (
            [[b"hello", True], [b"world", False]],
            ["bytes", "bool"],
            "db6cebb0ef6e70da569bb9b581f4044e143345e2e17f4afe069b334d3c3b87ab",
        ),
        # Test case 10: All data types combined and root
        (
            [
                [
                    i,
                    f"0xd3cda913deb6f67967b99d67acdfa1712c29360{i%10}",
                    b"bytesdata" + bytes(i),
                    bool(i % 2),
                ]
                for i in range(123)
            ],
            ["uint256", "address", "bytes", "bool"],
            "20b98c773d1f6a25963777358400ac38e0f63406b8e6e6d078fca9c4a9920853",
        ),
    ],
)
def test_root_merkle_tree(leaves, types, root):
    standartMerkleTree = StandartMerkleTree()

    standartMerkleTree.of(leaves, types)

    assert standartMerkleTree.root == root


def test_new_tree():

    standartMerkleTree = StandartMerkleTree()

    leaves = [[0], [1], [2], [3], [4]]
    types = ["int"]

    tree = standartMerkleTree.of(leaves, types)

    assert tree._tree == [
        None,
        "2c2ba3fc1b5a2d1d6bb3d96aea9c1eeb5ebab4a4fecd3278dd0c9eb098ce9862",
        "e97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162",
        "c167b0e3c82238f4f2d1a50a8b3a44f96311d77b148c30dc0ef863e1a060dcb6",
        "a3c45fa76377c0a96e373793e9023fb2a2d6970e0989ca39ce9f50d9acf834c9",
        "673620737675e2755ce8269a99904022d15da8d5843f5aec205cd243ff80240a",
        "c167b0e3c82238f4f2d1a50a8b3a44f96311d77b148c30dc0ef863e1a060dcb6",
        None,
        "510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9",
        "b5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22",
        "1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0",
        "2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01",
        "c167b0e3c82238f4f2d1a50a8b3a44f96311d77b148c30dc0ef863e1a060dcb6",
        None,
        None,
        None,
    ]
    assert tree._format == "standard-v1"
    assert (
        tree._root
        == "2c2ba3fc1b5a2d1d6bb3d96aea9c1eeb5ebab4a4fecd3278dd0c9eb098ce9862"
    )

    assert (
        tree.root
        == "0x2c2ba3fc1b5a2d1d6bb3d96aea9c1eeb5ebab4a4fecd3278dd0c9eb098ce9862"
    )

    assert tree._tree == standartMerkleTree._tree


def test_get_proofs():

    standartMerkleTree = StandartMerkleTree()

    leaves = [[0], [1], [2], [3], [4]]
    types = ["int"]

    standartMerkleTree.of(leaves, types)

    leaf_hash1 = standartMerkleTree.leafToHash(leaf=[0], types=["int"])
    leaf_hash2 = standartMerkleTree.leafToHash(leaf=[1], types=["int"])
    leaf_hash3 = standartMerkleTree.leafToHash(leaf=[2], types=["int"])
    leaf_hash4 = standartMerkleTree.leafToHash(leaf=[3], types=["int"])

    assert (
        leaf_hash1
        == "510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9"
    )
    assert (
        leaf_hash2
        == "b5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22"
    )
    assert (
        leaf_hash3
        == "1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0"
    )
    assert (
        leaf_hash4
        == "2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01"
    )

    assert standartMerkleTree.getProofs(leaf=leaf_hash1) == [
        "0xb5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22",
        "0x673620737675e2755ce8269a99904022d15da8d5843f5aec205cd243ff80240a",
        "0xc167b0e3c82238f4f2d1a50a8b3a44f96311d77b148c30dc0ef863e1a060dcb6",
    ]
    assert standartMerkleTree.getProofs(leaf=leaf_hash2) == [
        "0x510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9",
        "0xc167b0e3c82238f4f2d1a50a8b3a44f96311d77b148c30dc0ef863e1a060dcb6",
        "0x2c2ba3fc1b5a2d1d6bb3d96aea9c1eeb5ebab4a4fecd3278dd0c9eb098ce9862",
    ]
    assert standartMerkleTree.getProofs(leaf=leaf_hash3) == [
        "0x2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01",
        "0xc167b0e3c82238f4f2d1a50a8b3a44f96311d77b148c30dc0ef863e1a060dcb6",
        "0xc167b0e3c82238f4f2d1a50a8b3a44f96311d77b148c30dc0ef863e1a060dcb6",
    ]
    assert standartMerkleTree.getProofs(leaf=leaf_hash4) == [
        "0x1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0",
        "0xa3c45fa76377c0a96e373793e9023fb2a2d6970e0989ca39ce9f50d9acf834c9",
        "0x2c2ba3fc1b5a2d1d6bb3d96aea9c1eeb5ebab4a4fecd3278dd0c9eb098ce9862",
    ]


def test_invalid_get_proofs():

    standartMerkleTree = StandartMerkleTree()

    leaves = [[0], [1], [2], [3], [4]]
    types = ["int"]

    standartMerkleTree.of(leaves, types)

    with pytest.raises(Exception) as exc_info:
        standartMerkleTree.getProofs(leaf="some_leaf")

    assert str(exc_info.value) == "some_leaf is not a leaf"


def test_properties():

    standartMerkleTree = StandartMerkleTree()

    leaves = [[0], [1], [2], [3]]
    types = ["int"]

    standartMerkleTree.of(leaves, types)

    assert (
        standartMerkleTree.root
        == "e97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162"
    )
    assert standartMerkleTree.generated == True
    assert standartMerkleTree.numberOfNodes == 4
    assert standartMerkleTree.tree == [
        "e97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162",
        "a3c45fa76377c0a96e373793e9023fb2a2d6970e0989ca39ce9f50d9acf834c9",
        "673620737675e2755ce8269a99904022d15da8d5843f5aec205cd243ff80240a",
        "510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9",
        "b5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22",
        "1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0",
        "2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01",
    ]
    assert standartMerkleTree.sizeOfTree == 8
    assert standartMerkleTree.depth == 3

    assert standartMerkleTree.leaves == [
        "510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9",
        "b5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22",
        "1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0",
        "2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01",
    ]


def test_reset_the_tree():

    standartMerkleTree = StandartMerkleTree()

    leaves = [[0], [1], [2], [3]]
    types = ["int"]

    standartMerkleTree.of(leaves, types)

    # Reset
    standartMerkleTree.reset()

    assert standartMerkleTree.root == None
    assert standartMerkleTree.generated == False
    assert standartMerkleTree.numberOfNodes == 0
    assert standartMerkleTree.tree == []
    assert standartMerkleTree.sizeOfTree == 0
    assert standartMerkleTree.depth == 0
    assert standartMerkleTree.leaves == []


def test_generate_with_str_bool():

    standartMerkleTree1 = StandartMerkleTree()
    standartMerkleTree2 = StandartMerkleTree()
    standartMerkleTree3 = StandartMerkleTree()

    types = ["bool"]

    leaves1 = [[True], [False]]
    leaves2 = [["true"], ["false"]]
    leaves3 = [["True"], ["False"]]

    standartMerkleTree1.of(leaves1, types)
    standartMerkleTree2.of(leaves2, types)
    standartMerkleTree3.of(leaves3, types)

    assert (
        standartMerkleTree1.root
        == standartMerkleTree2.root
        == standartMerkleTree3.root
    )


def test_generate_with_str_bytes():

    standartMerkleTree1 = StandartMerkleTree()
    standartMerkleTree2 = StandartMerkleTree()
    standartMerkleTree3 = StandartMerkleTree()

    types = ["bytes"]
    # bytes string like 'aaaaaaaa'
    leaves1 = [
        [b"\xaa\xaa\xaa\xaa"],
        [b"\xbb\xbb\xbb\xbb"],
        [b"\xcc\xcc\xcc\xcc"],
        [b"\xdd\xdd\xdd\xdd"],
        [b"\xee\xee\xee\xee"],
        [b"\xff\xff\xff\xff"],
    ]

    # hex string like '0xaaaaaaaa'
    leaves2 = [[f"0x{i*8}"] for i in ["a", "b", "c", "d", "e", "f"]]

    # string like aaaaaaaa
    leaves3 = [[f"{i*8}"] for i in ["a", "b", "c", "d", "e", "f"]]

    standartMerkleTree1.of(leaves1, types)
    standartMerkleTree2.of(leaves2, types)
    standartMerkleTree3.of(leaves3, types)

    assert (
        standartMerkleTree1.root
        == standartMerkleTree2.root
        == standartMerkleTree3.root
    )


def test_concat():

    result = StandartMerkleTree.concat("0xAAA", "0xBBB")
    assert result == "0xAAABBB"

    result = StandartMerkleTree.concat("0xAAA", "0xBBB", sort=True)
    assert result == "0xAAABBB"

    result = StandartMerkleTree.concat("0xAAA", "0xBBB", sort=False)
    assert result == "0xAAABBB"

    result = StandartMerkleTree.concat("", "")
    assert result == "0x"

    result = StandartMerkleTree.concat("0xAAA", "")
    assert result == "0xAAA"

    result = StandartMerkleTree.concat("123", "456", sort=True)
    assert result == "0x123456"

    result = StandartMerkleTree.concat("aBc", "cBa", sort=True)
    assert result == "0xaBccBa"


def test_invalid_concat():

    with pytest.raises(ValueError, match="concat function works with strings"):
        StandartMerkleTree.concat(123, "0xBBB")

    with pytest.raises(ValueError, match="concat function works with strings"):
        StandartMerkleTree.concat("0xAAA", 456)


def test_print_tree(capsys):

    standartMerkleTree = StandartMerkleTree()

    leaves = [[0], [1], [2], [3]]
    types = ["int"]

    standartMerkleTree.of(leaves, types)

    # Call the function that prints to stdout
    standartMerkleTree.printTree()

    # Capture the printed output
    captured = capsys.readouterr()

    # Check the printed output
    assert (
        captured.out
        == "============== TREE ==============\n1 |  |_: e97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162\n2 |  |  |__: a3c45fa76377c0a96e373793e9023fb2a2d6970e0989ca39ce9f50d9acf834c9\n4 |  |  |  |___: 510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9\n5 |  |  |  |___: b5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22\n3 |  |  |__: 673620737675e2755ce8269a99904022d15da8d5843f5aec205cd243ff80240a\n6 |  |  |  |___: 1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0\n7 |  |  |  |___: 2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01\n"
    )


def test_str():

    standartMerkleTree = StandartMerkleTree()
    leaves = [[0], [1], [2], [3]]
    types = ["int"]

    standartMerkleTree.of(leaves, types)
    # Call the function that prints to stdout

    assert (
        str(standartMerkleTree)
        == "Tree Root Hash: e97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162"
    )

    # Empty-Tree

    standartMerkleTree.reset()

    assert str(standartMerkleTree) == "The tree has no leaves yet."


def test_is_valid_hex_string_raises_value_error():
    tree = StandartMerkleTree()

    with pytest.raises(ValueError, match="is not hex-string."):
        tree.of(values=[["0xInvalid"]], types=["bytes"])

    with pytest.raises(ValueError, match="is not hex-string."):
        tree.of(values=[["0xNotHex"]], types=["bytes"])

    with pytest.raises(ValueError, match="is not hex-string."):
        tree.of(values=[["0x0x0x"]], types=["bytes"])

    with pytest.raises(ValueError, match="is not hex-string."):
        tree.of(values=[["0x12345z"]], types=["bytes"])

    with pytest.raises(ValueError, match="is not hex-string."):
        tree.of(values=[["0z0x"]], types=["bytes"])

    with pytest.raises(ValueError, match="is not hex-string."):
        tree.of(values=[["12345z"]], types=["bytes"])


def test_is_valid_integer():
    tree = StandartMerkleTree()

    with pytest.raises(ValueError, match="is not integer or unsigned integer."):
        tree.of(values=[["abc"]], types=["uint"])

    with pytest.raises(ValueError, match="is not integer or unsigned integer."):
        tree.of(values=[["12.34"]], types=["uint256"])

    with pytest.raises(ValueError, match="is not integer or unsigned integer."):
        tree.of(values=[["0x12"]], types=["int"])

    with pytest.raises(ValueError, match="is not integer or unsigned integer."):
        tree.of(values=[["None"]], types=["uint256"])


def test_is_valid_boolean():
    tree = StandartMerkleTree()

    with pytest.raises(ValueError, match="is not boolean."):
        tree.of(values=[["abc"]], types=["bool"])

    with pytest.raises(ValueError, match="is not boolean."):
        tree.of(values=[["TRUE"]], types=["bool"])

    with pytest.raises(ValueError, match="is not boolean."):
        tree.of(values=[["FALSE"]], types=["bool"])

    with pytest.raises(ValueError, match="is not boolean."):
        tree.of(values=[["None"]], types=["bool"])


def test_check_bytes_invalid():
    tree = StandartMerkleTree()

    with pytest.raises(ValueError, match="is not bytes."):
        tree.of(values=[[[0x01, 0x02, 0x03]]], types=["bytes"])

    with pytest.raises(ValueError, match="is not bytes."):
        tree.of(values=[[123]], types=["bytes32"])

    with pytest.raises(ValueError, match="is not bytes."):
        tree.of(values=[[True]], types=["bytes32"])


#################################################
#################### Tree #######################
#################################################


def test_tree_init():

    standartMerkleTree = StandartMerkleTree()
    leaves = [[0], [1], [2], [3]]
    types = ["int"]

    tree = standartMerkleTree.of(leaves, types)

    assert tree._types == ["int"]
    assert tree._values == [[0], [1], [2], [3]]
    assert tree._format == "standard-v1"
    assert (
        tree._root
        == "e97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162"
    )

    assert tree._tree == [
        None,
        "e97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162",
        "a3c45fa76377c0a96e373793e9023fb2a2d6970e0989ca39ce9f50d9acf834c9",
        "673620737675e2755ce8269a99904022d15da8d5843f5aec205cd243ff80240a",
        "510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9",
        "b5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22",
        "1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0",
        "2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01",
    ]

    assert (
        repr(tree)
        == "MerkleTree(\n\tformat: 'standard-v1',\n\ttree:\n\t  '0xe97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162'\n\t  '0xa3c45fa76377c0a96e373793e9023fb2a2d6970e0989ca39ce9f50d9acf834c9'\n\t  '0x673620737675e2755ce8269a99904022d15da8d5843f5aec205cd243ff80240a'\n\t  '0x510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9'\n\t  '0xb5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22'\n\t  '0x1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0'\n\t  '0x2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01',\n\tvalues:\n\t\t[[0], [1], [2], [3]],\n\troot:\n\t  0xe97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162\n\tleaf_encoding:  ['int'])"
    )


def test_tree_properties():
    standartMerkleTree = StandartMerkleTree()
    leaves = [[0], [1], [2], [3]]
    types = ["int"]

    tree = standartMerkleTree.of(leaves, types)

    assert tree.leaf_encodings == ["int"]
    assert tree.values == [[0], [1], [2], [3]]
    assert tree.format == "standard-v1"
    assert (
        tree.root
        == "0xe97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162"
    )

    assert tree.tree == [
        "e97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162",
        "a3c45fa76377c0a96e373793e9023fb2a2d6970e0989ca39ce9f50d9acf834c9",
        "673620737675e2755ce8269a99904022d15da8d5843f5aec205cd243ff80240a",
        "510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9",
        "b5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22",
        "1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0",
        "2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01",
    ]


def test_tree_hexify():

    standartMerkleTree = StandartMerkleTree()
    leaves = [[0], [1], [2], [3]]
    types = ["int"]

    tree = standartMerkleTree.of(leaves, types)

    assert tree.tree == [
        "e97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162",
        "a3c45fa76377c0a96e373793e9023fb2a2d6970e0989ca39ce9f50d9acf834c9",
        "673620737675e2755ce8269a99904022d15da8d5843f5aec205cd243ff80240a",
        "510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9",
        "b5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22",
        "1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0",
        "2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01",
    ]

    new_tree = tree._tree_hexify(tree.tree)

    assert new_tree == [
        "0xe97e7870ffec941070df3ac0255521c04008ba2c5a2b9ebb33e0d4adee758162",
        "0xa3c45fa76377c0a96e373793e9023fb2a2d6970e0989ca39ce9f50d9acf834c9",
        "0x673620737675e2755ce8269a99904022d15da8d5843f5aec205cd243ff80240a",
        "0x510e4e770828ddbf7f7b00ab00a9f6adaf81c0dc9cc85f1f8249c256942d61d9",
        "0xb5d9d894133a730aa651ef62d26b0ffa846233c74177a591a4a896adfda97d22",
        "0x1ab0c6948a275349ae45a06aad66a8bd65ac18074615d53676c09b67809099e0",
        "0x2584db4a68aa8b172f70bc04e2e74541617c003374de6eb4b295e823e5beab01",
    ]



from dotenv import dotenv_values
import sys
import os
import pytest
import inspect
import re

from geodefi.utils.merkle import StandartMerkleTree
from geodefi import Geode
from geodefi.classes.validator import Validator


sys.path.append(os.getcwd())


env = dotenv_values(".env")


G = Geode(exec_api=env["EXECUTION_API"], cons_key=env["CONSENSUS_KEY"])
PORTAL = G.Portal


POOL_LIST = [
    "50016835115526216130031110555486827201953559012021267556883950029143900999178",
    "29228457249232120346521013786824808088246537603535847808963148138747123868265",
    "13237608342217427718393841084329878006268509356659527629195599913782184459845",
    "70609728262553005376169614685188798579261852324203356795003385284436695946570",
    "84581558291237928112291510488374176467358729712668559583760486858349467929555",
]


def test_init():

    myPool = PORTAL.pool(POOL_LIST[0])

    # == 50016835115526216130031110555486827201953559012021267556883950029143900999178
    assert isinstance(myPool.ID, int)
    assert myPool.TYPE == 5
    assert myPool.network == 5

    assert myPool.Beacon == G.Beacon
    assert myPool.w3 == G.w3

    myPool2 = PORTAL.pool(
        50016835115526216130031110555486827201953559012021267556883950029143900999178
    )
    assert myPool.ID == myPool2.ID


def test_invalid_init():

    # with pytest.raises(ValueError, match=re.escape("invalid literal for int() with base 10: 'string'")):
    #    PORTAL.pool('string')

    with pytest.raises(
        TypeError,
        match=re.escape(
            "int() argument must be a string, a bytes-like object or a real number, not 'list'"
        ),
    ):
        PORTAL.pool(["LIST", "LII"])


# WARNING: readAddress is not working....

# assert myPool.maintainer == 0
# assert myPool.middlewaresList == 0
# assert myPool.whitelist == 0
# assert myPool.withdrawalContract == 0
# assert myPool.liquidityPool == 0


def test_allowance():
    myPool = PORTAL.pool(POOL_LIST[0])

    operatorID = 114391297015478800753082638170652680401082080549997516459063441314156612391510
    # 114391297015478800753082638170652680401082080549997516459063441314156612391510

    assert myPool.allowance(operatorID) == 10

    assert myPool.allowance(0) == 0

    assert (
        myPool.allowance(
            106440257855610140438147021303249733137151692982873052796064618175630395030085
        )
        == 0
    )

    with pytest.raises(
        TypeError,
        match=re.escape("The following abi value is not a 'uint256': -1"),
    ):
        myPool.allowance(-1) == 0

    with pytest.raises(
        TypeError,
        match=re.escape("The following abi value is not a 'uint256': 11.4"),
    ):
        myPool.allowance(11.4) == 0


def test_proposedValidators():
    myPool = PORTAL.pool(POOL_LIST[0])

    operatorID = 114391297015478800753082638170652680401082080549997516459063441314156612391510
    # 114391297015478800753082638170652680401082080549997516459063441314156612391510

    assert myPool.proposedValidators(operatorID) == 0

    assert (
        myPool.proposedValidators(
            106440257855610140438147021303249733137151692982873052796064618175630395030085
        )
        == 0
    )

    with pytest.raises(
        TypeError,
        match=re.escape("The following abi value is not a 'uint256': -1"),
    ):
        myPool.proposedValidators(-1) == 0

    with pytest.raises(
        TypeError,
        match=re.escape("The following abi value is not a 'uint256': 11.4"),
    ):
        myPool.proposedValidators(11.4) == 0


def test_activeValidators():
    myPool = PORTAL.pool(POOL_LIST[0])

    operatorID = 114391297015478800753082638170652680401082080549997516459063441314156612391510
    # 114391297015478800753082638170652680401082080549997516459063441314156612391510

    assert myPool.activeValidators(operatorID) == 1

    assert (
        myPool.activeValidators(
            106440257855610140438147021303249733137151692982873052796064618175630395030085
        )
        == 0
    )

    with pytest.raises(
        TypeError,
        match=re.escape("The following abi value is not a 'uint256': -1"),
    ):
        myPool.activeValidators(-1) == 0

    with pytest.raises(
        TypeError,
        match=re.escape("The following abi value is not a 'uint256': 11.4"),
    ):
        myPool.activeValidators(11.4) == 0


# WARNING: readAddress does not work...

"""
def test_middlewares():
    myPool = PORTAL.pool(POOL_LIST[0])

    print(myPool.middlewares(0))
    print(myPool.middlewares(1))
    print(myPool.middlewares(-1))
"""


def test_validators():

    myPool = PORTAL.pool(POOL_LIST[0])

    # Call the validators method
    myVal = myPool.validators(0)

    # Check if the returned object is an instance of Validator
    assert isinstance(myVal, Validator)

    # Check if the attributes of the Validator object match the expected values
    assert myVal.network == G.network
    assert myVal.beacon == G.Beacon
    assert myVal.w3 == G.w3
    assert myVal.poolId == int(POOL_LIST[0])


def test_prepareProposeStake():
    pass


def test_prepareBeaconStake():
    pass


myPool = PORTAL.pool(POOL_LIST[0])


def test_initiated():
    assert myPool.initiated == 1677379164


def test_surplus():
    assert myPool.surplus == 0


def test_secured():
    assert myPool.secured == 0


def test_middlewaresLen():
    assert myPool.middlewaresLen == 1


def test_private():
    assert myPool.private == 0


def test_withdrawalCredential():
    assert (
        myPool.withdrawalCredential
        == "0x010000000000000000000000c82ed5ec571673e6b18c4b092c9cbc4ae86c786e"
    )


def test_feeSwitch():
    assert myPool.feeSwitch == 0


def test_priorFee():
    assert myPool.priorFee == 0


def test_fee():
    assert myPool.fee == 500000000


def test_wallet():
    assert myPool.wallet == 0


def test_validatorsLen():
    assert myPool.validatorsLen == 1


def test_validatorsList():

    assert myPool.validatorsList == [
        b"\x93&\xf6\xc0\x7f\x8a\xbd\x08.\xf8+\x19'\x9c\xbb\xa7ak\x03\x95\xfb\x94}P\xcd-_\xef0=\xd6\x13\xab\xe3\x10\x87\x07zg\xfa\xa4w\xc0c\x1c\xc7\"\x8d"
    ]



import pytest
import sys

from pytest import MonkeyPatch
from geodefi.exceptions import PythonVersionError
from geodefi.geodefi import check_python_version


def test_python_invalid_version_check():
    # Mock sys.version_info to return a specific version (3.6 in this case)
    with pytest.raises(PythonVersionError):
        with pytest.raises(SystemExit):
            with MonkeyPatch.context() as m:
                m.setattr(sys, "version_info", (3, 6))
                check_python_version()


def test_python_version_check():
    # Mock sys.version_info to return a version greater than 3.7 (e.g., 3.8)
    with MonkeyPatch.context() as m:
        m.setattr(sys, "version_info", (3, 8))
        check_python_version()
        assert True
